using namespace std;
#include <iostream>
#include <iomanip>
#include <sstream>
#include <unistd.h>
#include "calout.h"
#include "probe.h"
#include <syslog.h>
#include "turbid.h"
#include "sinpi.h"
#include "tprops.h"
#include "alsa_dsp.h"

#ifdef TUNER
#  include <sys/ioctl.h>	/* for ioctl, TCGETA */
#  include <termios.h>		/* for ECHO */
#endif

extern "C" {
// This should have been provided by unistd.h but it wasn't;
// I have no idea why not:
       pid_t getpgid(pid_t pid);
};

datum smash(double foo) {
  if (foo > INT_MAX) return INT_MAX;
  if (foo < -INT_MAX) return -INT_MAX;
  return datum(foo);
}

double ofreq;
#ifdef TUNER

int fn = fileno(stdin);
// successive half-steps:
string kbd("\t1q2we4r5t6yu8i9op-[=]");
string keymap("C+D+EF+G+A+BC");
const double halfstep(pow(2.0, 1.0/12.0));
int octave(4);
int kbkey(9);		// A4 = 440 Hz
int abskey(57);

void dokey(int newkey){
  abskey = newkey;
  ofreq = 440*pow(halfstep, abskey-57);
  int myoct = abskey / 12;
  int mykey = abskey - myoct*12;
  string keyname;
  char keylet = keymap[mykey];
  if (keylet == '+') {
    keyname += keymap[mykey-1];
    keyname += '#';
  } else keyname += keylet;

  fprintf(stderr, "\r\033[K %2s%1d %6.2f\r", keyname.c_str(), myoct, ofreq);
  fflush(stderr);
}

int escmode(0);

void chkey(){
  char cc;
  int nn = read(fn, &cc, 1);
  if (nn <= 0) return;
  if (cc == 033) {
    escmode = 1;
    return;
  }
  if (escmode) {
    if (cc == '~')  escmode = 0;
    if (cc != 'O' && isupper(cc)) escmode = 0;
    if (cc == 'B' || cc == 'D') dokey(abskey-1);
    else if (cc == 'A' || cc == 'C') dokey(abskey+1);
    return;
  }
  string::size_type where = kbd.find(cc);
  if (where != kbd.npos) {
    kbkey = where;
  } else if (cc == ',') {
    octave--;
    if (octave < 0) octave = 0;
  } else if (cc == '.') {
    octave++;
    if (octave > 8) octave = 8;
  } else {
    int foo;
    foo = cc;
    fprintf(stderr, "%04o\n", foo);
    return;
  }
  dokey(12*octave + kbkey);
}

void setkey(){
  kbd += char(127);	// the "backspace" key
  kbd += '\\';
  kbd += char(10);	// the "return" key
}

#else
inline void chkey(){}		// does nothing
inline void setkey(){}		// does nothing
#endif

// print, in a fixed width, with comma
template<typename T> void p(T const foo){
  cout << setw(12) << foo << ",";
}

// print within quote marks, in a fixed field width, with comma
template<typename T> void pq(T const foo){
// the stringstream is required because we want the field
// _including quote marks_  to have the specified width.
  stringstream xxx("");
  xxx << "\"" << foo << "\"";
  cout << setw(12) << xxx.str() << ",";
}

vector<pollfd> calibrator::pollist() {
  vector<pollfd> pollme;
  if (!cal_dsp.phndl) return pollme;
  int cnt = snd_pcm_poll_descriptors_count(cal_dsp.phndl);
  pollme.resize(cnt);
  snd_pcm_poll_descriptors(cal_dsp.phndl, &pollme[0], pollme.size());
  return pollme;
}

int calibrator::setup(){
  if (!cal_dsp.phndl) return -1;
  int somecal(0);
  int needfreq(0);
  if (tpx.sine_level > -500) {
    somecal++;
    needfreq++;
  }
  if (tpx.square_level > -500) {
    somecal++;
    needfreq++;
  }
  if (tpx.broad_level > -500) {
    somecal++;
    /* no frequency needed */
  }
  if (tpx.comb_level > -500) {
    somecal++;
    needfreq++;
  }

  if (!somecal) {
    cerr << "No amplitude???" << endl;
    return(-1);	// no amplitude: don't even bother.
  }

// some modes require a frequency:

  if (needfreq && !tpx.calfrq.size()) {
    cerr << "No frequency??" << endl;
    return(-1);	// no amplitude: don't even bother.
  }

  setkey();

  activeNsig = cal_dsp.channels.size();

// A smaller output block size
// makes things noticeably less efficient
// (In contrast, the input block size has little if any
//  effect on efficiency.)
  calblock__frames = cal_dsp.intperiod__fr * 4;

  if (show["calout"]) {

    cerr << "Calout starts";
    if (tpx.calfrq.size())
      cerr << "  ofreq: " << tpx.calfrq[0];
    else
      cerr << "  (no frequency) ";
    cerr << "  hwNchan: " << cal_dsp.hwNchan
        << "  activeNsig: " << activeNsig
        << endl
        << "  INT_MAX: " << INT_MAX
        << "  sizeof(datum): " << sizeof(datum)
        << "  calblock__frames: " << calblock__frames
        << endl;
    if (0) for (unsigned int ii = 0; ii < cal_dsp.channels.size(); ii++) {
       cout << "calout: signal: " << ii
            << " on hw channel: " << cal_dsp.channels[ii]
            << " freq: " << tpx.calfrq[ii]
            << endl;
    }
  }


#ifdef TUNER
  {
    termio save;
     // If ioctl fails, we're probably not connected to a terminal.
    int rslt;
    rslt = ioctl(fn, TCGETA, &save);
    termio t(save);
    t.c_lflag &= ~ECHO;
    t.c_lflag &= ~ICANON;
    t.c_cc[VMIN] = 1;
    t.c_cc[VTIME] = 0;
    t.c_cc[VERASE] = 0;

    ioctl(fn, TCSETA, &t);
    int temp(1);
    ioctl(fn, FIONBIO, &temp);
  }
#endif

  calout_block.resize(calblock__frames * cal_dsp.hwNchan);

// 1: int_ref_level is easiest to understand for square wave.
// 2: dB is a measure of power, not voltage.
// Therefore a sine wave peaks sqrt(2) higher than int_ref_level
  sinamp = int_ref_level * pow(10, tpx.sine_level/20) * sqrt(2.);
  squamp = int_ref_level * pow(10, tpx.square_level/20);
  broad_factor = pow(10, tpx.broad_level/20);
  comb_amp = int_ref_level * pow(10, tpx.comb_level/20) * sqrt(2.);

// FIXME:  broad_phase probably should be a vector indexed by signal:
  broad_phase = 0;   // integer phase, for pseudo-random probe signal

// each rotor is essentially a static variable within this process:
// initialize it before starting the main loop:
// ditto for each comb_rotor
  sine_rotor.resize(activeNsig);

  nyquist = cal_dsp.rate / 2.;

// implement a whole bunch of sine waves:
// The comb_offset is close to a half, and its
// denominator (D) is close to a prime number.
// The 75th harmonic will collide, but I don't care about that.
// 75 == 2*D - 1
        double comb_offset = 18. / 37.;
  comb_interval = tpx.calfrq[0];
  nspike = floor(nyquist / comb_interval);
  comb_rotor.resize(activeNsig);

// initialize all the rotors:
  for (unsigned int sig=0; sig < activeNsig; sig++){
    sine_rotor[sig].init(tpx.calfrq[sig] / cal_dsp.rate);
    comb_rotor[sig].resize(nspike);
    for (int spk = 0; spk < nspike; spk++) {
      double afreq = comb_interval * (comb_offset + spk);
      comb_rotor[sig][spk].init(afreq / cal_dsp.rate);
    }
  }
  if (show["comb"]) {
    cout << "Comb: " << nspike << " spikes, namely: ";
    for (int spk = 0; spk < nspike; spk++){
      double afreq = comb_interval * (comb_offset + spk);
      cout << setw(13) << afreq << ",";
    }
    cout << endl;
  }
  return 0;
}

void calibrator::io_step(){
  if (!cal_dsp.phndl) return;

   if (1) {		// do one block

// loop over all frames in this block,
// generating data:
    for (unsigned int ff = 0; ff < calblock__frames; ff++) {
      chkey();

// zeroize all hw channels in frame, whether or not they are active:
      ptr = &calout_block[ff * cal_dsp.hwNchan];
      for (unsigned int jj = 0; jj < cal_dsp.hwNchan; jj++) {
	*ptr++ = 0;
      }

// loop over all _active_ signals in frame:
      ptr = &calout_block[ff * cal_dsp.hwNchan];

      for (unsigned int asig = 0; asig < activeNsig; asig++){
        double raw(0);
//notused//        double local_90;

// implement broadband hiss:
        raw += broad_factor * probe[broad_phase];
        broad_phase++;
        if (broad_phase >= nprobe) broad_phase = 0;

// implement simple sine wave:
        raw += sinamp * sine_rotor[asig].acc.imag();
        if (0) cout << "xx: (" << sine_rotor[asig].acc.real()
                << ", " << sine_rotor[asig].acc.imag() << ")"
                << " + (" << sine_rotor[asig].c_step.real()
                << ", " << sine_rotor[asig].c_step.imag() << ")"
        << endl;
// phase per frame for this signal
        ++sine_rotor[asig];
        /* sine wave and square wave share the phase variable */

// implement square wave:
        double reduced_phase = sine_rotor[asig].subphase;
        raw += reduced_phase >= 0.5 ? squamp : -squamp;
        /* sine wave and square wave share the phase variable */

// implement comb:
        if (tpx.comb_level > -500.) for (int spk = 0; spk < nspike; spk++){
          raw += comb_amp * comb_rotor[asig][spk].acc.imag();
          ++comb_rotor[asig][spk];
        }

// see where to find this signal in the hardware block:
        int hwch = cal_dsp.channels[asig];
        ptr[hwch] = smash(raw);
      }
    }
// end of generating data

// get a pointer to the start of the block:
    datum* doit(&calout_block[0]);
    int todo = calblock__frames;

    while (todo) {		// keep trying to get this block out
      int rslt = keep.writei(doit, todo);
      if (rslt == todo) {
        //////// cerr << "FYI: calout good write" << endl;
        break;		// good!
      }
      if (rslt > 0) {			// short write
        // ALSA docs say this will never happen
        // unless there is a signal or an underrun
        // http://www.alsa-project.org/alsa-doc/alsa-lib/group___p_c_m.html
        /////// cerr << "FYI: calout short write" << endl;
        todo -= rslt;
	doit += rslt;
	continue;
      }
      if (rslt == -EPIPE) {		// explicit underrrun
        if (show["underrun"]) syslog(LOG_WARNING,
		"Turbid calout underrun: wrote %d got %d: %s",
                todo, rslt, snd_strerror(rslt));
        snd_pcm_prepare(cal_dsp.phndl);	// recover from underruns
	continue;
      }
      // some error we don't understand
      syslog(LOG_WARNING, "Turbid calout wrote %d got %d", todo, rslt);
      snd_pcm_prepare(cal_dsp.phndl);	// try to recover
    } // end trying to get this block out
    oblock++;
  }
}

// A Keeper maintains an i/o buffer and information about
// its state, such as how full it is.
// Constructor
Keeper::Keeper(alsa_dsp* _dsp, int const _fpp)
  : dsp(_dsp),                  // remember the args
    fpp(_fpp),
    nchan(_dsp->hwNchan),
    buf(nchan * _fpp),     // allocate the buffer
    finbuf(0)
{}

// Just like the ALSA library routine of the same name, except:
// 1) The first arg is a "keeper",
//   which means that we will accept writes that are
//   not an integer multiple of the period.
// 2) The second arg is a datum* (not void*).
//   We call writei() to convert things to the sample-size
//   of the actual device.
snd_pcm_sframes_t Keeper::writei(datum const* buffer,
        snd_pcm_uframes_t appbuf__fr){
  int todo(appbuf__fr);
  int head(0);          // frames taken from head of buffer
  const datum* mybuf(buffer);
  datum* start = &buf[0];
  if (finbuf) {
    head = fpp - finbuf;
    memcpy(start + nchan*finbuf, buffer, nchan*sizeof(datum)*head);
    int frames_read = dsp->writei(start, fpp);
    if (frames_read <= 0) return frames_read;         // no change in kk
    if (frames_read != fpp) return -98;
    todo -= head;
    finbuf = 0;
  }
  mybuf += nchan*head;
  int rslt = dsp->writei(mybuf, todo);
  if (rslt < 0) return rslt;    // some error, let caller deal with it
  if (rslt == 0) return head;           // head got written, nothing more
// OK, at this point rslt must be positive.
  int left(todo - rslt);                // number of frames left to do....
  if (!left) return appbuf__fr;           // Nothing?  Great!
  if (left < fpp) {          // can we keep it for next time?
    finbuf = left;
    memcpy(start, mybuf + nchan*rslt, nchan*sizeof(datum)*left);
    return appbuf__fr;
  }
// here with a short write that doesn't fit in our keep-buffer:
  finbuf = 0;
  return head + rslt;
}
