//////////////////////////////////////////////////////////////////////
//
// code to cleanly detach from the controlling terminal
// aka the "daemon fork".
//
// This seems kinda complicated to me ... but I've seen lots of
// simpler approaches that didn't work reliably.

using namespace std;

#include <iostream>		/* needed by some compilers */
#include <sys/types.h>		/* for fork, for wait */
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>		/* for wait */
#include <sys/resource.h>	/* for wait */
#include <sys/wait.h>
#include <sys/stat.h>		/* for umask */

#include "turbid.h"
#include "daemon_fork.h"
#include <string.h>             /* for strerror */

daemon_fork::daemon_fork() :
  parent_pid(0)
{}

const int earlyclose(3);	// protect stdin, stdout, stderr

// Setup MUST be separate from the constructor.
// A fork inside a constructor is asking for trouble.
void daemon_fork::setup(){
  parent_pid = getpid();
  int kidpid = fork();  
  switch (kidpid) {
  case -1:
    cerr << "Couldn't fork." << endl;
    exeunt(2);
  case 0:
    break;	// We are the child 
  default:	// We are the parent
    int status;
    for (;;) {
      int rslt;
      rslt = wait4(kidpid, &status, WUNTRACED, 0);
// Should never get here, unless worker process exits
// before killing us during its detach().
// Try to analyze what happened:
      if (rslt > 0) break;
      if (errno == EINTR) continue;
      if (errno == ECHILD) break;	// KLUDGE!!!
      cerr << "wait4 kidpid " << kidpid
      	<< " failed " << rslt 
	<< " -- " << strerror(errno) << endl;
      exeunt(1);
    }
// if it exited, let's hope it printed a message;
// we will just exit with the same status:
    if (WIFEXITED(status)) exit(WEXITSTATUS(status));

// Hmmm, hard to understand, just say what we know:
    cerr <<  "Daemon process died early: " << status
	<< " = " << (status >> 8) << " " << (status & 255) << endl;

    _exit(0);	
  }

  if (0) for (int fd = earlyclose; fd<getdtablesize(); fd++){
    close(fd);
  }
  // child just returns;  will need to call detach() later....
}

#include <syslog.h>


void daemon_fork::detach(){
  if (!parent_pid) {
    return;	// possible and reasonable, if setup() was never called
    		// or detach() was previously called
  }
// child code:
  setsid();

  chdir("/");

  umask(077);

  if (verbosity) cerr << "Child about to detach: " << getpid() << endl;
  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);

  int rslt = kill(parent_pid, SIGUSR1);
  if (rslt) syslog(LOG_ERR, "Error killing parent %d: %m ", parent_pid);
  parent_pid = 0;	// so detach is a no-op if called again
}
