/////////////////////////////////////////////////////////////////
// Harvest entropy from audio I/O system.

// In part inspired by audio-entropyd.c by Damien Miller

#if 0

Official online documentation
  http://www.alsa-project.org/alsa-doc/alsa-lib/

Info about HDA-Intel:
  http://www.intel.com/support/motherboards/desktop/sb/cs-015851.htm


#endif

using namespace std;
#include <iostream>
#include <iomanip>
#include <sstream>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <alsa/asoundlib.h>
#include <vector>
#include "math.h"
#include <string>
#include <vector>
#include <map>

#ifdef BOGUS_DECLARE_GETPGID
extern "C" {
// This should have be provided by unistd.h ... but in some
// releases it isn't; I have no idea why it isn't
// (or wasn't) on some platforms:
       pid_t getpgid(pid_t pid);
};
#endif /* BOGUS_DECLARE_GETPGID */
#include <syslog.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <asm/types.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>
#include <string>
#include <vector>
#include "minmax.h"
#include <vector>
#include <linux/random.h>       /* for ioctl RNDGETENTCNT */
#include <poll.h>

// On debian, you need to install the development package,
//  i.e. libasound2-dev (not just libasound2) to provide
//  alsa/asoundlib.h and the associated runtime libraries:
#include <alsa/asoundlib.h>
// undocumentd alsa function:

#include "gauss.h"
#include "dbuffer.h"
#include "stretch.h"
#include "logger.h"

#include "turbid.h"

#include "utils.h"              /* for less_casefold */
//#include "alsa_dsp.h"         /* included LATER in this file */


// some kludgely global variables
char const * colorized;

#include "defs4turbid.h"


// Note that in the \turbid{} software, anything that is linear in
//  voltage is represented relative to 1.0 = full scale.  In contrast,
//  anything that is measured in dB is measured relative to a reference
//  level that is 15 dB below full scale.  This means that a rail-to-rail
//  square wave (such as might result from extreme clipping) is reported
//  as $+15$ dB. The peak voltage of a zero-dB sine wave $\sqrt{}2$
//  higher than a zero-dB square wave.  Here's a numerical example of
//  what you might see at this stage:

class alsa_dsp;         // forward reference

#include "tprops.h"
#include "alsa_dsp.h"

#include "daemon_fork.h"
#include "alsactl.h"
#include "calout.h"
#include "turbid.h"
#include "probe.h"
#include "arg_parser.h"
#include "sinpi.h"
#include "riff-wav.h"
#include "utils.h"              /* for getTime() */

class captor{
public:
  tprops& tpx;
  int hwNchan;
  int ringlen;
  int ringptr;
  vector<datum> cap_buf;
  alsa_dsp * cap_dsp;

// Constructor:
  captor(tprops& _foo, alsa_dsp &in)
  : tpx(_foo), cap_dsp(&in)
  {}
  void setup();
  void io_step();
  vector<pollfd> pollist();
};

typedef responder_base * rsb_p;

// Design note:  Wrappers are appropriate when a single
// object might have multiple FDs open.  We can then put
// multiple copies of the wrapper into the responder map,
// without having multiple copies of the underlying object.

// The wrapper
template<typename T>
  class wrapper : public responder_base{
public:
// In addition to what's in the responder base,
// we need to keep track of "our" object:
  T& object;

// Constructor;  tell the base what fd and events we care about:
  wrapper<T>(pollfd const xx, T& ours)
  : object(ours)
  {
    fd = xx.fd;
    events = xx.events;
  }

// wrap "our" object's i/o routine:
  virtual void io_step() {
    object.io_step();
  }
};

vector<pollfd> captor::pollist(){
  int cnt = snd_pcm_poll_descriptors_count(cap_dsp->phndl);
  vector<pollfd> pollme(cnt);
  snd_pcm_poll_descriptors(cap_dsp->phndl, &pollme[0], pollme.size());
  return pollme;
}

// STDIN doesn't need a wrapper
class stdin_resp : public responder_base{
public:
  vector<char> buf;
  stdin_resp(int const _fd, short _events){
    fd = _fd;
    events = _events;
  }
  virtual void io_step() {
    buf.resize(10000);
    int rslt = read(0, &buf[0], buf.size());
    if (rslt < 0) {
      cerr << "stdin read error: ";
      perror(0);
      exit(1);
    }
    cout << "+++" << string(&buf[0], rslt);
  }
  vector<pollfd> pollist() {
    vector<pollfd> rslt(1);
    rslt.front().fd = fd;
    rslt.front().events = events;
    return rslt;
  }
};


// a few non-constant global variables:
string pidfile;
int verbosity=0;
int daemonflag=0;
double tzero(0);
shower show;
string progname;

typedef long long int llint;

// Forward references:

int get_cardno(const string justcard);
void config_mixer(tprops& tpx, int const cardno);
void write_pid_file(const string fname);
void dummy(int signum);
void trap_exit(int signum);
void trap_detach(int signum);
void drain(char const* draindev);
void drain(int const fd);
int isfile(char const* fname);
string tolower(const std::string& arg);
void colorize(ostream& ouch, char const * color);
void decolorize(ostream& ouch, char const * color);

// Functions

int& shower::operator[](string const check){
  if (!topics.count(check)) {
    cerr << "Warning:  shower never heard of '" << check << "'" << endl;
  }
  return topics[check];
}

void shower::touch(initializer_list<string> const _arg){
  list<string> arg(_arg);
  for (list<string>::const_iterator aptr = arg.begin();
            aptr != arg.end(); aptr++) {
    string tpx(*aptr);
// If the entry already exists, this does nothing:
    topics.insert(pair<string,int>(tpx,0));
  }
}

void tprops::usage(int const err) {
  (err ? cerr : cout) <<
"Harvest entropy from the audio I/O system.\n"
"Usage: " << progname << " [options]\n\n"
"Examples:\n"
"  " << progname << " freq 440 ochan 0   # play sound on left channel\n"
"  " << progname << " freq 440 ochan 1   # play sound on right channel\n"
"  " << progname << " freq 440 ochan 0,1 # play perfect 4th, using both channels\n"
"  " << progname << " @tpx.tbd           # take commands from file\n"
"  " << progname << " show args          # show args and parameters\n"

"Options:\n"
"@             []  Take commands from file\n"
"card          []  Sound card, subdevice to use; default hw:0,0\n"
"mixer-ctl     []  Load mixer setup from alsactl file.\n"
"                '' ==> don't set up; just use inherited settings.\n"
"amplitude     []  Calibrator amplitude / dB; cumulative, if given more than once.\n"
"ichan         <>  List of  input channels, e.g. 0,1 ; ichan -1 means none\n"
"ochan         <>  List of output channels;  -1 means none; default: all channels\n"
"frequency     []  Calibrator 1st chan freg, in Hz; hint: 440.\n"
"interval      []  Calibrator 2nd chan freq, relative to 1st; default 4./3.\n"

"frame-rate    []  Audio measurement rate, frames per second;\n"
"                 0 ==> default ==> max rate supported by the hardware.\n"
"\n"
"Vopen         []  Voltmeter reading, open circuit, for calibration.\n"
"Vloaded       []  Voltmeter reading, loaded by Rload, for calibration.\n"
"Rload         []  Resistor for measuring Vloaded, for calibration.\n"
"cnet          <>  List of series/shunt/series/shunt resistors.\n"
"point         <>  Previous calibration observation\n"
"lbw           []  Lockin bandwidth in Hz;  default 10.\n"
"Qrep          []  Quantum of representation, from calibration.\n"
"Rin           []  Input resistance, from calibration.\n"
"\n"
"detach            Detach from terminal (daemonize).\n"
"lurk              Lurk (instead of exiting) if device busy.\n"
"pidfile-dir   []  File to contain our PID.\n"
"output-fifo   []  Main random output fifo filename.\n"
"stretch-fifo  []  Stretched-random fifo.\n"
"bandwidth     []  Bandwidth / Hz\n"
"\n"
"verbose           Print extra debugging info (verb verb ==> even more).\n"
"show          []  show info on specified topic.\n"
"show topics       list of available topics for 'show' command.\n"
"sps               (dangerous kludge) directly specify entropy per sample.\n"
"wfile         []  Write raw audio data to specified file, .wav format\n"
/////"           \"\" ==> default ==> don't bother.\n"
"snooze        []  Time to wait between displaying calibration result lines;\n"
"                      has no effect if writing checkfile or principal fifo.\n"
"exit              Exit at end of main loop, if wfile done.\n"
"\n"
"Hint: use the companion program 'randomize' to feed entropy to kernel.\n"

<< endl;
}

// Constructor
tprops::tprops() :
  card_device("hw:0,0"),
  argXXXichan(0), explicit_ichan(0),
  argXXXochan(0), explicit_ochan(0),
  rando_fname("/dev/hrandom"),
  stretch_fname("/dev/srandom"),
  pid_fname("/var/run/" + progname + ".pid"),
  wfile_fname(""),
  wfile_count(0),
// Note that for the hdaintel driver, the /alsa period/ is 1028
// and the /alsa buffer/ is 4096,
// neither of which is equal to our chosen /application buffer/
// but this seems to be OK.
  exit_flag(0),
  desired_rate(0),
  XVout(0),
  XVin(0),
  bandwidth(0),
  Rout(0),
  Vopen(0),
  Vloaded(0),
  Rload(Inf),
  /* cnet  defaults to zero-length list */
  /* point defaults to zero-length list */
  qrep(0),                      // quantum of representation
  r_in(0),
  sine_level(-999),      // default amplitude (dB relative to -15dBFS)
  square_level(-999),
  broad_level(-999),
  comb_level(-999),
  argXXXofreq(440.),
  argXXXinterval(4./3.),
  lbw(10.),                     // lockin bandwidth
  lurking(0),
  mixer_ctlfile("default"),

  obs(512),     // there are two buffers, so actual storage is twice this
  rando_fd(-1),
  stretch_fd(-1),
  wfile_fd(-1),
  sps(0),
  start_time(0.),
  snooze(.5),
  last_printout(0),
  sincon(0)
{}      // no code, just the initializers (above)

// Reference:
// http://www.tldp.org/HOWTO/Bash-Prompt-HOWTO/x329.html
char const* crtRed   ("\033[1;31m");
char const* crtGreen ("\033[1;32m");
char const* crtBlue  ("\033[1;34m");
char const* crtNormal("\033[0;39m");

vector<int> clean_me;

void trapme(){
// FIXME: use sigaction instead
  signal(SIGHUP, trap_exit);            // hangup
  signal(SIGINT, trap_exit);            // interrupt
  signal(SIGTERM, trap_exit);           // terminate
  signal(SIGPIPE, trap_exit);           // broken pipe
  signal(SIGUSR1, trap_detach);         // daemon_fork kills parent
  signal(SIGCHLD, SIG_DFL);
}

// could conceivably be moved to utils.c:
double rms(int const* ptr, int const nn){
  if (!nn) return 0;
  double val;
  double sum(0);
  for (int ii = 0; ii < nn; ii++){
    val = ptr[ii];
    sum += val*val;
  }
  sum /= nn;
  return sqrt(sum);
}

string q(const string tpx){
  return "'" + tpx + "'";
}

void basic_setup(){
// static_assert requires compiling with -std=c++0x
  static_assert (sizeof(int) == 4, "Turbid assumes 4-byte integers");

  show.touch({"readcycle", "cpoint", "stuck", "lockin",
        "stats", "topics", "fdsetup", "args", "hwproperties",
        "hwsetup", "xvout", "rout", "calout", "comb",
        "clipping", "overrun", "underrun", "timing"});
}

int main(int const argc, char const * const * argv) {
  basic_setup();

  double probe_level(rms(probe, nprobe));
// fudge to account for the fact that the reference has
// no power at DC and no power at the Nyquist frequency:
  double fudge_ref = int_ref_level * sqrt((nprobe-2) / double(nprobe));
  if (abs(probe_level/fudge_ref - 1) > 1e-9) {
    cerr << "Mismatch: probe_level: " << probe_level
        << "  fudge_ref: " << fudge_ref
        << "  check: " << probe_level/fudge_ref - 1
        << " which should be smaller than 1e-9 (typically 3e-12)"
        << endl;
// should never happen, unless bogus probe signal is
// compiled in.
    exit(1);
  }

  progname = argv[0];
  string::size_type where = progname.rfind('/');        // find final slash
  if (where != progname.npos) {
    progname = progname.substr(1+where);
  }
  tprops tpx;                            // holder for cmdline options ...
  tpx.cmdline(argc, argv);              // ... process them
  tpx.iomap[0] = "stdin";
  tpx.iomap[1] = "stdout";
  tpx.iomap[2] = "stderr";
  tzero = getTime();
  tpx.set_defaults();
  tpx.show_topics();
  tpx.show_args();
  trapme();                             // install trap handlers
  daemon_fork myfork;
  if (daemonflag) myfork.setup();   // so we know the right pid
  write_pid_file(tpx.pid_fname);        // write it to file

  {
    int logflags = LOG_CONS | LOG_PID;
    if (!daemonflag) logflags |= LOG_PERROR;
    openlog(progname.c_str(), logflags, LOG_DAEMON);
  }

// Lurkers should detach early, so we don't hang the
// init.d/turbid startup script.
// Non-lurkers should not detach early, to increase the
// chance of detecting errors and exiting with a meaningful
// error status.

  if (tpx.lurking) myfork.detach();

// Be sure to setup the capture handler first, so
// on a soundblaster16 capture gets the 16-bit DMA
// channel and calout gets the 8-bit channel.
//
//   (In contrast, after setup, we want the input
//   stream to start /running/ before the output
//   stream, or (better) at the same time.)

  alsa_dsp in("capture");

// buf_serno gets incremented before any consumers get to see it;
// consumers see the first buffer having serno==0
  in.buf_serno = -1;
// Beware that messing with buf_serno
// as a means to skip frames is
// probably misleading and useless or worse
// if the input blocksize and output blocksize are different.
// Instead we need some kind of skip_frames.

// Probably shouldn't skip frames at all when writing
// to a file;  let postprocessor (e.g. wavetool) decide
// what it wants to do.

//??  in.skip_frames = 0;           // may get changed later

  int err;
  if (tpx.desired_rate >= -1) {
    err = in.setup(tpx.card_device,
          SND_PCM_STREAM_CAPTURE,
          tpx.explicit_ichan ? &tpx.argXXXichan : 0,
          tpx);
    if (err) exit(1);           // msgs have already been printed
    in.show_setup(tpx);
// arrange for 1 second of data:
    if (tpx.wfile_fname.length()) tpx.wfile_count = 1 * in.rate;
  }

  alsa_dsp out("calout");

  int somecal(0);
  {
    if (tpx.sine_level > -500) somecal++;
    if (tpx.square_level > -500) somecal++;
    if (tpx.broad_level > -500) somecal++;
    if (tpx.comb_level > -500) somecal++;
    double my_ofreq(tpx.argXXXofreq);

// Don't initialize the output device unless it is needed.
// It might be busy or otherwise unavailable, and we
// don't want to incur an error if we don't have to.
    if (somecal                         // off ==> don't even initialize
        || tpx.desired_rate < 0) {      // unless we are just dumping params FIXME
      err = out.setup(tpx.card_device,
            SND_PCM_STREAM_PLAYBACK,
            tpx.explicit_ochan ? &tpx.argXXXochan : 0,
            tpx, 2);
      if (err) exit(1);           // msgs have already been printed
#ifdef disgusting
      usleep(100000);             // let hardware settle down ...
            // (SB16 has nasty transients right after initialization)
#endif
// Set up the frequencies after initializing the device, so
// we know how many channels there actually are.
// Do it before invoking calout().

      tpx.calfrq.resize(out.channels.size());
      for (unsigned int ii = 0; ii < tpx.calfrq.size(); ii++) {
        tpx.calfrq[ii] = my_ofreq;
        my_ofreq *= tpx.argXXXinterval;
      }
      in.show_setup(tpx);
    }  // end if somecal or ...
  } // end scope of my_ofreq

// Be sure to configure mixer  _after_ opening the pcm handlers,
// because we want to hold exclusive access to the mixer (control
// device) but snd_pcm_open wants to temporarily open the control
// device for its own reasons.
  config_mixer(tpx, get_cardno(tpx.justcard));

  tpx.open_devices();           // random-output, raw-output, etc.

  if (verbosity) syslog(LOG_NOTICE, "starting %s", progname.c_str());
  myfork.detach();
// Now that detach has done the setsid, we can fork our subprocesses:

// ... the dbuffer process:
  if (tpx.sps && tpx.rando_fd >= 0)
    clean_me.push_back(dbuffer(tpx.bufpipe, tpx.rando_fd, tpx.obs));

// ... the stretch process:
  if (tpx.sps && tpx.stretch_fd >= 0)
    clean_me.push_back(stretch(tpx.bufpipe, tpx.stretch_fd, tpx.obs));

//
  if (somecal) {
    int rslt = snd_pcm_link(in.phndl, out.phndl);
    if (rslt < 0) {
      cerr << "FYI: failed to link the two ALSA streams\n";
    } else if (verbosity) {
      cerr << "Linked!\n";
    }
  }

  tpx.show_fdsetup(&in, &out);

// close the read end of bufpipe after all consumers [i.e. dbuffer, stretch]
// have been started.  This means the main program will be killed via
// SIGPIPE (rather than blocking) if/when the last consumer dies.
// Although dying is much better than blocking,
// it is an open question whether we should try to keep running,
// since there could be other output-related things going on,
// including user-interface things.
  close(tpx.bufpipe[0]);        // main program will never read this pipe

#if 0 /* hack hack hack */
// This is useful in special cases to show how to interpret timing plots,
// and also useful for showing the effect of snd_pcm_link().
// See ./timing.gnumeric and ../paper/turbid.htm for details
        usleep(10000);
#endif

#if 1   /* write the .wav header */
  if (tpx.wfile_count > 0) {
    WAV_header hdr;
    int hwNchan    = in.hwNchan;             // number of hw-supported channels

    hdr.setup(WAVE_FORMAT_PCM, in.rate, hwNchan,
                  32, double(tpx.wfile_count) / in.rate);

    int todo = sizeof(hdr);
    int rslt = write(tpx.wfile_fd, &hdr, todo);
    if (rslt != todo) {
      cerr << "Short write on WAV header: " << endl;
      perror(0);
      exit(1);
    }
  }
#else
  /* just dump the raw data with no .WAV header */
#endif

// Do some "easy" calculations, i.e. the ones that
// don't require any input data.
  tpx.show_xvout();
  tpx.show_rout();

// Get set up for main loop.
  captor cappy(tpx, in);
  cappy.setup();
  calibrator calout(tpx, out);  // harmless if "out" is inactive
  calout.setup();               // harmless if "out" is inactive

  typedef map<int,rsb_p> resp_t;
  resp_t resp;

  stdin_resp scanner(0, POLLIN);
  vector<pollfd> v = scanner.pollist();
  for (pollfd xx: scanner.pollist())
    resp[xx.fd] = &scanner;
  for (pollfd xx: cappy.pollist())
    resp[xx.fd] = new wrapper<captor>(xx, cappy);
  for (pollfd xx: calout.pollist())     // pollist is nil if "out" is inactive
    resp[xx.fd] = new wrapper<calibrator>(xx, calout);

// If linked, input device starts when output device does,
// on first write.  If not linked, we have to start it by
// hand, because if it's not running, ppoll will never see
// it as ready to read.
  if (!out.phndl) snd_pcm_start(in.phndl);

  vector<pollfd> pollme;
  for (auto & xx : resp) {
    pollme.push_back({xx.first, xx.second->events});
    cout << "hi there: " << xx.first << " = " << xx.second->fd << endl;
  }

  struct timespec timeout({5, 987654321});
  for (;;) {            // the main loop over events
    int ready = ppoll(&pollme[0], pollme.size(), &timeout, 0);
    if (ready < 0) {
      cerr << "ppoll failed : ";
      perror(0);
      exit(1);
    }
//xxx    cout << "ppoll returns: " << ready << endl;
    for (pollfd & xx : pollme){
      if (xx.revents) {
//xxx        cout << "invoking for fd " << xx.fd << endl;
        resp[xx.fd]->io_step();
      } else {
//xxx        cout << "nothing for fd " << xx.fd << endl;
      }
    }

  } // end main loop over events
  exeunt(0);
}  // end of int main

void decolorize(ostream& ouch, char const * color){
  if (color) {
    ouch << crtNormal;
  }
  colorized = 0;
}

void colorize(ostream& ouch, char const * color){
  colorized = color;
  if (color) {
    ouch << color;
  }
}

string purify(const string tpx) {
  string rslt;
  for (string::size_type ii = 0; ii < tpx.length(); ii++) {
    char ch = tpx[ii];
    if (isupper(ch)) ch += 'a' - 'A';
    if (ch == ' ') rslt += '_';
    else if (isalnum(ch)) rslt += ch;
  }
  return rslt;
}

// return a string like "sb16.ctl"
// based on ALSA's notion of the driver name
//
// justcard can be something like hw:0 (but not hw:0,0)
string invent_ctlfile(const string justcard, int const verbosity){
    string driver("unknown");
    snd_ctl_t* chndl;           // control handle
    int err = snd_ctl_open(&chndl, justcard.c_str(), SND_CTL_RDONLY);
    if (err < 0) {
      chndl = 0;
      cerr << "invent_ctlfile: error opening control channel to card "
          << justcard << endl;
      // proceed with driver = "unknown"
      // probably will run into fatal problems later ....
    } else {
      snd_ctl_card_info_t *info;
      snd_ctl_card_info_malloc(&info);
      err = snd_ctl_card_info(chndl, info);
      if (err < 0) {
        cerr << "Can't get card ID info: " << err << endl;
      } else {
        driver = snd_ctl_card_info_get_mixername(info);
      }
      snd_ctl_card_info_free(info);
    } // found a control channel
    return purify(driver) + ".ctl";
}

// convert card-ID string to integer card index
int get_cardno(const string justcard){
    snd_ctl_t* chndl;           // control handle
    int err = snd_ctl_open(&chndl, justcard.c_str(), SND_CTL_RDONLY);
    if (err < 0) return -1;
    snd_ctl_card_info_t *info;
    snd_ctl_card_info_malloc(&info);
    err = snd_ctl_card_info(chndl, info);
    if (err < 0) return -2;
    int rslt = snd_ctl_card_info_get_card(info);
    snd_ctl_card_info_free(info);
    return rslt;
}

//static string digits = "0123456789";

// Process command-line switches, et cetera
void tprops::cmdline(int const argc, char const * const * argv) {

  show["underrun"]++;        // turn this on by default
  show["overrun"]++;        // turn this on by default

// Note basic default values were set up in constructor.
// Now process options:


  arg_parser arg(argc, argv);
  arg.fold_case = 1;
  string progname = arg.next();
  string raw_arg, raw_raw_arg;
  string hackhack;
  string fname;

  try {for (;;){
    raw_raw_arg = arg.next();         // get next keyword
    if (arg.fail()) break;
// silently strip leading '-' or '--'
    if (arg.cur[0] == '-') arg.cur = arg.cur.substr(1);
    if (arg.cur[0] == '-') arg.cur = arg.cur.substr(1);
    raw_arg = arg.cur;

    int where = arg.cur.length();
    if (where){
      if (arg.cur[where-1] == ':') arg.cur = arg.cur.substr(0,where-1);
    }
    if (0) {}
    else if (arg.prefix("@")) {arg >> fname; arg.push(fname); }
    else if (raw_arg[0] == '@') {arg.push(raw_arg.substr(1)); }
    else if (raw_arg == "==>") { /* ignore */ }
    else if (arg.prefix("card")) arg >> card_device;
    else if (arg.prefix("mixer-ctl")) arg >> mixer_ctlfile;
    else if (arg.prefix("output-fifo")) arg >> rando_fname;
    else if (arg.prefix("stretch-fifo")) arg >> stretch_fname;
    else if (arg.prefix("pidfile")) arg >> pid_fname;
    else if (arg.prefix("sine")) {
      if (sine_level < -500) sine_level = 0;
      double delta;
      arg >> delta;
      sine_level += delta;
    }
    else if (arg.prefix("square")) {
      if (square_level < -500) square_level = 0;
      double delta;
      arg >> delta;
      square_level += delta;
    }
    else if (arg.prefix("broadband")) {
      if (broad_level < -500) broad_level = 0;
      double delta;
      arg >> delta;
      broad_level += delta;
    }
    else if (arg.prefix("comb")) {
      if (comb_level < -500) comb_level = 0;
      double delta;
      arg >> delta;
      comb_level += delta;
    }
    else if (arg.prefix("ichan")) {
      arg >> argXXXichan;
      explicit_ichan++;
    }
    else if (arg.prefix("ochan")) {
      arg >> argXXXochan;
      explicit_ochan++;
    }
    else if (arg.prefix("bandwidth")) arg >> bandwidth;
    else if (arg.prefix("frequency")) arg >> argXXXofreq;
    else if (arg.prefix("interval")) arg >> argXXXinterval;
    else if (arg.prefix("lbw")) arg >> lbw;
    else if (arg.prefix("frame-rate")) arg >> desired_rate;
    else if (arg.prefix("Vopen")) arg >> Vopen;
    else if (arg.prefix("Vloaded")) arg >> Vloaded;
    else if (arg.prefix("Rload")) arg >> Rload;
    else if (arg.prefix("cnet")) arg >> cnet;
    else if (arg.prefix("cpoint")) arg >> cpoint;
    else if (arg.prefix("XVout")) arg >> XVout;
    else if (arg.prefix("XVin")) arg >> XVin;
    else if (arg.prefix("Rout")) arg >> Rout;
    else if (arg.prefix("Qrep")) arg >> qrep;
    else if (arg.prefix("Rin")) arg >> r_in;
    else if (arg.prefix("show")) {
      string args("");
      arg >> args;
      while (args.length()) {
        string::size_type where = args.find(',');
        string arg;
        if (where == args.npos) {
          arg = args;
          args = "";
        } else {
          arg = args.substr(0,where);
          args = args.substr(1+where);
        }
        if (arg[0] != '-') show[arg]++;
        else {
          arg = arg.substr(1);
          if (show[arg] > 0) show[arg]--;
        }
      }
    }
    else if (arg.prefix("verbosity")) { verbosity++; }
    else if (arg.prefix("detach")) {  daemonflag++; }
    else if (arg.prefix("lurk")) { lurking++; }
    else if (arg.prefix("wfile")) arg >> wfile_fname;
    else if (arg.prefix("snooze")) arg >> snooze;
    else if (arg.prefix("exit")) { exit_flag++; }
    else if (arg.prefix("help")
          || arg.prefix("-help")
          || arg.prefix("--help")) {
      usage(0);
      exit(0);
    }
    else if (arg.prefix("hack")) arg >> hackhack;
    else if (arg.prefix("sps")) arg >> sps;
    else {
      cerr << "Unrecognized keyword: '" << raw_arg << "'\n";
      cerr << "... for help, try\n  " << progname
             << " help" << endl;
      exit(1);
    }
  }} catch (string err){
    cerr << err << endl;
    exit(1);
  }
  time_t mytime(time(0));
  srand(mytime);

}

void tprops::set_defaults(){

// figure out default place to look for config file, pid file:
  char const* homish = getenv("HOME");
  if (!homish) homish = "/root";
  string home(homish);
  if (home != "/root") {
    pid_fname = home + pid_fname;
    rando_fname = home + rando_fname;
    stretch_fname = home + stretch_fname;
  }

  string::size_type where = card_device.find(',');
  if (where == card_device.npos) justcard = card_device;
  else justcard = card_device.substr(0, where);

#if 0
// snd_card_get_index does not work for names of the form "hw:0,0" nor "hw:0"
// nor "hw:CARD=Intel,DEV=0" nor "hw:CARD=Intel"
  cout <<  card_device << " :: "  << snd_card_get_index(card_device.c_str()) << endl;
  cout <<  justcard << " :: " << snd_card_get_index(justcard.c_str()) << endl;
#endif

  if (mixer_ctlfile == "default") {
    mixer_ctlfile = invent_ctlfile(justcard, verbosity);
  }// else cmdline has explict ctlfile, use that
  if (mixer_ctlfile.length()) {
    char mx = mixer_ctlfile[0];
    if (mx != '.' && mx != '/') {
      mixer_ctlfile = "/etc/" + progname + "/" + mixer_ctlfile;
      if (home != "/root") {
        mixer_ctlfile = home + mixer_ctlfile;
      }
    } // else . or / means use it verbatim
  }

}

void tprops::show_topics(int const now){
  showerthing topics(show.getlist());
  if (now || show["topics"]) {
    cout << "Topics are:" << endl;
    for (map<string,int>::const_iterator tpx = topics.begin();
            tpx != topics.end(); tpx++){
      cout << "... " << setw(12) << tpx->first << ": " << tpx->second << endl;
    }
  }
}// display_topics

static string decode_ppoll(short events){
  if (!events) return "";
  stringstream rslt;
#define X(code) if (events & code) rslt << "|" << #code; \
                events &= ~code;
  X(POLLIN)
  X(POLLPRI)
  X(POLLOUT)
  X(POLLERR)
  X(POLLHUP)
  X(POLLNVAL)
#undef X
  if (events) rslt << "|XX" << oct << events;
  string temp = rslt.str();
  if (temp.length()) temp = temp.substr(1);
  return temp;
}

void tprops::show_fdsetup(alsa_dsp * in, alsa_dsp * out, int const now){
  if (now || show["fdsetup"]) {
    for (int ii = 0; ii<getdtablesize(); ii++){
      int rslt = fcntl(ii, F_GETFL);
      if (rslt != -1) {
        cout << "FD " << setw(3) << ii << " flags "
                << oct << setw(7) << rslt << dec;
        if (iomap.count(ii)) {
          cout << " == " << iomap[ii];
        }
        cout << endl;
      } else if (0) {
        cout << "FD " << setw(3) << ii << " errno " << errno
                    << " == " << strerror(errno) << endl;
      }
    }
    cout << "........." << endl;
    for (alsa_dsp const * dsp : {in, out}) {
      if (dsp->phndl) {
        int cnt = snd_pcm_poll_descriptors_count(dsp->phndl);
        vector<pollfd> pollme(cnt);
        int rslt = snd_pcm_poll_descriptors(dsp->phndl,
                &pollme[0], pollme.size());
        if (rslt) {/* ??? */}
        for (int ii = 0; ii < cnt; ii++){
          cout << "   " << setw(10) << dsp->moniker
                << "  fd: " << pollme[ii].fd
                << "  events: " << oct << pollme[ii].events << dec
                << " = " << decode_ppoll(pollme[ii].events)
                << "  revents: " << oct << pollme[ii].revents << dec
                << " = " << decode_ppoll(pollme[ii].revents)
                << endl;
        }
      }
    }
  }
}

void tprops::show_args(int const now){
  if (now || show["args"]) {
    cout << "card:         " << card_device << endl;
    cout << "mixer-ctl:    " << mixer_ctlfile << endl;
    cout << "frame-rate:   " << desired_rate << " Hz desired" << endl;
    cout << "ichan:        (";
      for (unsigned int ii = 0; ii < argXXXichan.size(); ii++) {
        if(ii) cout << ", ";
        cout << argXXXichan[ii];
      }
      cout << ")" << endl;

    cout << "ochan:        (";
      for (unsigned int ii = 0; ii < argXXXochan.size(); ii++) {
        if(ii) cout << ", ";
        cout << argXXXochan[ii];
      }
      cout << ")" << endl;

    cout << "sine:         " << sine_level << " dB" << endl;
    cout << "square:       " << square_level << " dB" << endl;
    cout << "broadband:    " << broad_level << " dB" << endl;
    cout << "comb:         " << comb_level << " dB" << endl;
    if (calfrq.size())
      cout << "frequency:    " << calfrq[0] << " Hz" << endl;
    if (calfrq.size() > 1)
      cout << "interval:     " << calfrq[1]/calfrq[0] << endl;
    cout << endl;

    cout << "output-fifo:  " << q(rando_fname)   << endl;
    cout << "stretch-fifo: " << q(stretch_fname) << endl;
    cout << "pidfile-dir:  " << q(pid_fname)     << endl;
    cout << "wfile:        " << q(wfile_fname)  << endl;
    cout << "exit_flag:    " << exit_flag << endl;
    cout << endl;

    cout << "Vopen:        " << Vopen << endl;
    cout << "XVout:        " << XVout << endl;

    cout << "Vloaded:      " << Vloaded << endl;
    cout << "Rout:         " << Rout << endl;
    cout << "Rload:        " << Rload << endl;
    cout << "cnet:         (";
      for (unsigned int ii = 0; ii < cnet.size(); ii++) {
        if(ii) cout << ", ";
        cout << cnet[ii];
      }
      cout << ")" << endl;

    cout << "cpoint:       (";
      for (unsigned int ii = 0; ii < cpoint.size(); ii++) {
        if(ii) cout << ", ";
        cout << cpoint[ii];
      }
      cout << ")" << endl;

    cout << "XVin:         " << XVin << endl;

    cout << "lbw:          " << lbw << endl;
    cout << "Bandwidth:    " << bandwidth << endl;
    cout << "Rin:          " << r_in << endl;
    cout << "Qrep:         " << qrep << endl;
// FIXME:  What if we are putting out something
// other than a sine wave?
    if (XVout) {
      double gfactor(pow(10, sine_level/20));
  // The open-circuit RMS voltage we think we are putting out is:
      cout << "V_oc should be: " << gfactor * XVout << endl;
    }
  }
}

void tprops::weird_stuff(){
  double qsig(0);       // FIXME!

  if (qsig && r_in && bandwidth) {
     double sigmaJ(sqrt(4 * kT * r_in * bandwidth));
     entropy_digitized_gaussian ggg;
     sps = ggg(sigmaJ/qsig);
     if (verbosity) cout << "SigmaJ: " << sigmaJ
        << " Bits per sample: " << sps << endl;
  }
}

void config_mixer(tprops& tpx, int const cardno){
  string ctlfilestr = tpx.mixer_ctlfile;
  if (!ctlfilestr.length()) return;

  char const* ctlfile = ctlfilestr.c_str();
  int rslt = open(ctlfile, O_RDONLY);
  if (rslt < 0) {
    syslog(LOG_WARNING, "Could not open alsactl file '%s': %m "
        "... will use inherited mixer settings.", ctlfile);
    return;
  }
  tpx.iomap[rslt] = ctlfilestr;
  close(rslt);

  rslt = load_state(ctlfile, cardno, 1, 1);

  if (rslt) {
    syslog(LOG_ERR, "error in load_state");
    exit(1);
  }
}

// Open a named fifo, creating it if necessary.
// Used for purposes, including the main output i.e. /dev/hrandom
int open_fifo(tprops& tpx, const string fname) {
  if (!fname.length()) return -1;

  int rslt = mkfifo(fname.c_str(), 0644);
  if (rslt < 0) {
    if (errno != EEXIST) {
      syslog(LOG_ERR, "Couldn't create output fifo (%s): %m",
        fname.c_str());
      exit(1);
    } else {            // already exists

      struct stat statbuf;
      rslt = stat(fname.c_str(), &statbuf);
      if (rslt < 0) {
        syslog(LOG_ERR, "Couldn't stat output fifo (%s): %m",
                fname.c_str());
        exit(1);
      }
      if (!S_ISFIFO(statbuf.st_mode)){
        syslog(LOG_WARNING, "Output (%s) exists but isn't a fifo,"
                " hope you know what you're doing.",
                fname.c_str());
        // proceed
      }
    }
  }
// Here with named fifo existing in the filesystem.

// We (main process) quasi-open the named fifo for reading.
// Otherwise the open (for writing) would block for lack
// of a listener, which would be bad.
// Note that we never actually do any reading, so holding
// it open doesn't interfere with other processes (e.g.
// independent customers) who do the real reading.
  int junk = open(fname.c_str(), O_RDONLY | O_NONBLOCK);
  if (junk < 0) {
    syslog(LOG_WARNING, "failed to open fifo '%s' -- %m",
        fname.c_str());
    // don't exit -- maybe not fatal -- but weird.
  }
  tpx.iomap[junk] = fname + ".deblock";

  int fd = open(fname.c_str(), O_WRONLY);
  if (fd < 0) {
      syslog(LOG_WARNING, "Couldn't open fifo (%s) for writing: %m",
              fname.c_str());
      syslog(LOG_WARNING, "Proceeding in 'test' mode....");
  }
  tpx.iomap[fd] = fname + ".write";
  return fd;
}

// Open miscellaneous i/o devices:
void tprops::open_devices(){
  if (wfile_fname.length()) {
    wfile_fd = open(wfile_fname.c_str(), O_WRONLY|O_CREAT|O_TRUNC, 0600);
    if (wfile_fd < 0) {
      syslog(LOG_ERR, "Couldn't open checkfile for writing: %m");
      exit(1);
    }
    iomap[wfile_fd] = wfile_fname;
  }

  if (rando_fname.length()) {
    rando_fd = open_fifo(*this, rando_fname);
    /* open_fifo does error checking */
    /* open_fifo sets iomap */

// Also open the double-buffering pipe:
// (also contingent on having a rando_fname)
    int rslt = pipe(bufpipe);
    if (rslt < 0) {
      syslog(LOG_WARNING, "Can't open pipe for double-buffering: %m");
      exit(1);
    }
    iomap[bufpipe[0]] = "bufpipe.read";
    iomap[bufpipe[1]] = "bufpipe.write";
  }

  if (stretch_fname.length()) {
    stretch_fd = open_fifo(*this, stretch_fname);
    /* open_fifo does error checking */
    /* open_fifo sets iomap */
    if (0) cerr << "opened stretch '" << stretch_fname
      << "'  as fd " << stretch_fd
      << endl;
  }
}

class histogram {
public:
  double max_ent;
  int stuck;
  double max_possible_ent;

double step(const datum* buf, int const bufsize, int const slots,
         const datum mean, int const stride, int const loss,
        int& inrange);

double doit(const datum* buf, int const bufsize, int const slots,
         const datum mean, int const stride, int const sticky);

};

// inrange set to 1 iff no clipping occurs
//
double histogram::step(const datum* buf, int const bufsize, int const slots,
         const datum mean, int const stride, int const loss,
        int& inrange){
  max_possible_ent = log2(slots);
  inrange = 0;
  int histo[slots];
  for (int jj = 0; jj < slots; jj++){
    histo[jj] = 0;
  }
  datum xx;
  const datum* from = buf;
  for (int ii = 0; ii < bufsize; ii++){
    xx = *from;
    from += stride;
    xx /= loss;
    xx += slots/2;      // the middle slot corresponds to zero voltage
    if (xx < 0) xx = 0;
    if (xx >= slots) xx = slots-1;
    histo[xx] += 1;
  }

// OK, the histogram is built;  now evalute p log p.

  double sumlog(0);
  int nj;
// Note:
// - sum[p log p] = - sum[(ni/d) log (ni/d)] where d = sum[ni]
// =        - (sum[ni/d log ni] - sum[ni/d log d])
// =        (-1/d) (sum[ni log ni] - sum[ni log d])
// = log d - (1/d)  sum[ni log ni]
  for (int jj = 0; jj < slots; jj++){
    nj = histo[jj];
    if(nj) {
      sumlog += nj * log2(nj);
    }
  }
  if (show["sumlog"])
    printf("log2(bufsize) %g, sumlog %g, sumlog/bufsize %g, loss %d\n",
        log2(bufsize), sumlog, sumlog / bufsize, loss);
  if (histo[0] == 0 && histo[slots-1] == 0) inrange = 1;
  return log2(bufsize) - sumlog / bufsize;
}

// Calculate histogram.
// This higher-level routine iterates checking for stuck bits.
//
// stuck is a divisor, e.g. 1 (no bad bits) or 512 (9 bad bits)

double histogram::doit(const datum* buf, int const bufsize, int const slots,
         const datum mean, int const stride, int const sticky){
  max_ent = -1.;
  double ent;
  int inrange;
  stuck = 1;
  int trystick = sticky;
  do {
    ent = step(buf, bufsize, slots, mean, stride, trystick, inrange);
    if (ent > max_ent) {
      max_ent = ent;
      stuck = trystick;
    }
    if (verbosity > 1)
      fprintf(stderr, "Using trystick %d, histo returns %g max %g\n",
        trystick, ent, max_ent);
    trystick *= 2;
  } while (!inrange);
  return max_ent;
}

// These calculations don't even require any input data
void tprops::show_xvout(){
  char* color(0);
  if(!show["xvout"]) return;

// FIXME:  What if we are putting out something
// other than a sine wave?
  double gfactor(pow(10, sine_level/20));

// The RMS voltage we think we are putting out is
//      Vopen = gfactor * XVout
// So:  XVout = Vopen / gfactor

// calculate XVout based on Vopen

  if (Rload != Inf || Vloaded != 0.) {
    cerr << "Vopen with Rload and/or Vloaded makes no sense" << endl;
    exeunt(1);
  }

  double new_XVout(Vopen / gfactor);
  colorize(cout, color);
  cout << "XVout: " << new_XVout;
  cout << "  \t#  "
          << "  freq: " << calfrq[0]
          << "  sine_level: " << sine_level
          << "  Vopen: " << Vopen;

  if (new_XVout != XVout) {
    if (XVout != 0) cout << "  #was: " << XVout;
    XVout = new_XVout;
  }
  cout << crtNormal << endl;
}

void tprops::show_rout(){
// Calculate Rout given Rload and Vloaded
// Requires previously-determined XVout

// FIXME:  What if we are putting out something
// other than a sine wave?
  double gfactor(pow(10, sine_level/20));
  char* color(0);

  if (!show["rout"]) return;
  if (Rload == Inf) {
    cerr << "Can't calculate Rout without Rload" << endl;
    exeunt(1);
  }
  if (!Vloaded) {
    cerr << "Can't calculate Rload without Vloaded" << endl;
    exeunt(1);
  }
  double Voc = gfactor * XVout;
  // slope = (1/Vloaded - 1/Voc) / (1/Rload - 1/Inf)
  // Rout = slope * Voc
  //      = (Voc/Vloaded - 1) / (1/Rload)
  double new_Rout = (Voc/Vloaded - 1.) * Rload;
  colorize(cout, color);
  cout    << "Rout: " << new_Rout;
  cout    << "  \t#"
          << "  Rload: " << Rload
          << "  Vloaded: " << Vloaded
          << "  sine_level: " << sine_level
          ;;;
#if 0
  if (new_Rout != Rout) {
    if (Rout != 0) cout << "  #was: " << Rout;
    Rout = new_Rout;
  }
#endif
  cout << crtNormal << endl;
}

// Perform lockin analysis just one channel, for show["lockin"]
// Assume chdx has already been checked to be in bounds
// for /both/ input and calout
int tprops::lockin_chan(alsa_dsp* cap_dsp,
   datum* cap_buffer,
   int const which,             // index into calfrq
   int const chdx,              // hw index into the buffer
   int const stride,
   char const* chname,          // user-friendly name, e.g. "left"
   char const* color){

  int didprint = 0;

  double phpf = calfrq[which] / cap_dsp->rate;      // phase per frame for this signal
  if (!show["lockin"] && !show["cpoint"]) return didprint;
  double rcvd_00, rcvd_90;
  for(int frame = 0; frame < block__fr; frame++) {
    double val = double(cap_buffer[frame * stride + chdx]) / INT_MAX;
    rcvd_00 = lock_00[which](val * sinpi(2. * cap_phase[which]));
    rcvd_90 = lock_90[which](val * cospi(2. * cap_phase[which]));
    cap_phase[which] += phpf;
  }

  if (getTime() - last_printout >= snooze) {

// Assuming there is a sine wave in the input, we just
// calculated /half/ of its true amplitude.
// For compatibility with other measurements,
// we want to report the RMS [not the true amplitude]
// so we scale things up by sqrt(2)  [not the full 2]

    rcvd_00 *= sqrt(2.);
    rcvd_90 *= sqrt(2.);
    double bothsq = rcvd_00*rcvd_00 + rcvd_90*rcvd_90;
    double both = sqrt(bothsq);
    double dbpow(-999);
    if (both != 0.0) dbpow = 20 * log10(both / norm_ref_level);

    didprint = 1;
    colorize(cout, color);
    cout << setw(7) <<  "Lockin"
        << setprecision(4)
        << setw(6) << " 00: " << setw(10) << rcvd_00
        << setw(6) << " 90: " << setw(10) << rcvd_90
        << setw(8) << "  "  << setw(10) << ""
        << setw(7) << " both: " << setw(10) << both << " ibu"
        << setw(4) << " == "      << setw(10) << dbpow << " dBx";

//xx        << "  frq: " << calfrq[which]
//xx        << "  phpf: " << phpf
//xx        << "  phase: " << cap_phase[which];

    cout << crtNormal << endl;
    // switch back to black BEFORE printing each newline;
    // otherwise bad things happen when piping to /bin/head

    if (!show["cpoint"]) return didprint;
    double runtime = getTime() - start_time;
// wait until the filter has settled within one part in a thousand:
    if (runtime * lbw > 7.) {
      if (cnet.size() && XVout && Rout) {

// FIXME:  What if we are putting out something
// other than a sine wave?
        double gfactor(pow(10, sine_level/20));
    // The open-circuit RMS voltage we think we are putting out is:
        double Voc = gfactor * XVout;
        double Reff(Rout);
        for (unsigned int ii = 0; ii < cnet.size(); ii++) {
          if ((ii & 1) == 0) {  // series
            Reff += cnet[ii];
          } else {              // shunt
            Voc *= cnet[ii] / (Reff + cnet[ii]);
            Reff = Reff * cnet[ii] / (Reff + cnet[ii]);
          }
        }
        cout << setprecision(4)
             << "cpoint: " << Reff
             << "," << Voc/both
             << "   # Voc: " << Voc
             << "  Xi: " << both
             << "  sine_level: " << sine_level
             << "  cnet: (";
        for (unsigned int ii = 0; ii < cnet.size(); ii++) {
          if(ii) cout << ", ";
          cout << cnet[ii];
        }
        cout << ")" << endl;
        if (cpoint.size()) {
          if (cpoint.size() != 2) {
            cerr << "cpoint should be a list of length 2" << endl;
            exeunt(1);
          }
          double slope = (Voc/both - cpoint[1]) / (Reff - cpoint[0]);
          double intercept = Voc/both - slope * Reff;
          double other_intercept = cpoint[1] - slope * cpoint[0]; // should be the same

// XVin is the voltage, from an ideal zero-impedance voltage source,
// that would produce 0 dBx in the input buffer
          double XVin = intercept * norm_ref_level;
          double Rin = intercept / slope;
          double Vin = Voc * Rin / (Rin + Reff);
          cout << "XVin: " << XVin
               << "   # Vin: " << Vin
               << "  slope: " << slope
               << "  intercept: " << intercept
               << "  other_intercept: " << other_intercept
               << "  sine_level: " << sine_level
               << endl;
          cout << "Rin: " << Rin
               << "   # Vin: " << Vin
               << "  slope: " << slope
               << "  intercept: " << intercept
               << "  other_intercept: " << other_intercept
               << "  sine_level: " << sine_level
               << endl;
        }
      } else {
        cout << "need more" << endl;
      }
    }
  }
  return didprint;
}

// Perform lockin analysis for multiple channels
// use at most the first two active channels;
// ignore any others.
int tprops::lockin(alsa_dsp* cap_dsp, datum* cap_buffer){
  int didprint(0);

  unsigned int usableNsig = min(calfrq.size(), cap_dsp->channels.size());
  if (!usableNsig) return didprint;      // nothing to do

  int stride = cap_dsp->hwNchan;        // number of hw channels in the buffer

// detect the first time we are called.
// initialize a bunch of stuff.
  if (cap_dsp->buf_serno == 0) {
    if (verbosity) cout << "lockin initializing:"
      << "  usableNsig: " << usableNsig
      << "  hwNchan: " << cap_dsp->hwNchan
      << "  intperiod__fr: " << cap_dsp->intperiod__fr
      << "  block__fr: " << block__fr
      << "  blocks/sec: " << cap_dsp->rate / block__fr
      << endl;
    start_time = getTime();
    cap_phase.resize(usableNsig);
    lock_00.resize(usableNsig);
    lock_90.resize(usableNsig);
    for (unsigned int ii = 0; ii < usableNsig; ii++) {
      cap_phase[ii] = 0;
      lock_00[ii].set_corner(2.*M_PI*lbw/cap_dsp->rate);
      lock_90[ii].set_corner(2.*M_PI*lbw/cap_dsp->rate);
    }
  }

// Color code:
// We use right=red, which is standard in audio wiring.
// Left is not 100% standardized, but let's use left=blue.
// Beware: this differs from navigation and aviation, where left=red and right=green.

  didprint |= lockin_chan(cap_dsp, cap_buffer, 0, cap_dsp->channels[0], stride, "left", crtBlue);

  if (usableNsig > 1) {
    didprint |=
      lockin_chan(cap_dsp, cap_buffer, 1, cap_dsp->channels[1], stride, "right", crtRed);
  }
  return didprint;
}


// Analyze just one channel, for show["stats"] and/or show["clipping"]
// Assume chdx has already been checked to be in bounds.
int tprops::anachan(alsa_dsp* dsp,
   datum* cap_buffer,
   int const stride,
   int const chdx,              // channel index
   char const* chname,          // user-friendly name, e.g. "left"
   char const* color){

  int didprint(0);
  if (getTime() - last_printout < snooze) return didprint;

  int topcount(0), botcount(0);
  double top, bot;

  double val, mom_1st(0), mom_2nd(0), plogp;

// should not be block__fr here;
// should implement ring buffer
  for(int frame = 0; frame < block__fr; frame++) {
    if (frame == 0) top = bot = cap_buffer[chdx]/INT_MAX;
    val = double(cap_buffer[frame * stride + chdx]) / INT_MAX;
    mom_1st += val;
    mom_2nd += val*val;
    if (val > top) {
      top = val;
      topcount = 1;
    } else if (val == top) topcount++;

    if (val < bot) {
      bot = val;
      botcount = 1;
    } else if (val == bot) botcount++;
  }

  mom_1st /= block__fr;          // mean
  mom_2nd /= block__fr;          // mean square
  double rms = sqrt(mom_2nd);
  double stdev = sqrt(mom_2nd - mom_1st*mom_1st);

  if (show["stats"]) {
    didprint = 1;
// Do the normalization denominators:
    int histoslots(4096);
// FIXME:  should use user-supplied estimate of stuck bits
    histogram histo;
// FIXME:  should use ring buffer here:
    plogp = histo.doit(&cap_buffer[0], block__fr, histoslots,
                        int(mom_1st), stride, 1);

    double dbdev(-999);
    if (stdev != 0.0) dbdev = 20 * log10(stdev / norm_ref_level);
    colorize(cout, color);
    cout  << setprecision(4)
          << setw(7) << chname
          << " 1st: " << setw(10) << mom_1st
          << " 2nd: " << setw(10) << mom_2nd
          << setw(7) << " RMS: " << setw(10) << rms
          << " stdev: " << setw(10) << stdev << " ibu"
          << " == " << setw(10) << dbdev << " dBx"
          << " plogp: " << setw(7) << plogp
          << " outof: " << setw(5) << histo.max_possible_ent
          << " logstuck: " << setw(5) << log2(histo.stuck)
          << crtNormal << endl;
  // switch back to black BEFORE printing the newline;
  // otherwise bad things happen when piping to /bin/head
  }

  if (show["clipping"]) {
    didprint = 1;
    colorize(cout, color);
  // What is this
    cout << " Input clip check: "
         << setprecision(4)
         << "top " << setw(3) << topcount << " times @ " << setw(9) << double(top) << " ibu   "
         << "bot " << setw(3) << botcount << " times @ " << setw(9) << double(bot) << " ibu"
         << endl;
    decolorize(cout, color);
  }
  return didprint;
}

int tprops::analyze(alsa_dsp* dsp, datum* cap_buffer){
  int didprint(0);
  vector<datum> sdiff_buffer(block__fr);
  int activeNsig = dsp->channels.size();    // number of active signals we care about
  if (!activeNsig) return didprint;
  int stride = dsp->hwNchan;            // number of hw channels in the buffer
  int left_idx = dsp->channels[0];
  int right_idx = left_idx;
  if (activeNsig > 1) right_idx = dsp->channels[1];
  // use at most the first two active channels;
  // ignore any others.

// Color code:
// We use right=red, which is standard in audio wiring.
// Left is not 100% standardized, but let's use left=blue.
// Beware: this differs from navigation and aviation, where left=red and right=green.

  didprint |= anachan(dsp, cap_buffer, stride, left_idx, "left", crtBlue);

  if (right_idx != left_idx) {
    didprint |= anachan(dsp, cap_buffer,  stride, right_idx, "right", crtRed);
    vector<datum> sdiff_buffer(block__fr);
    for(int frame = 0; frame < block__fr; frame++) {
      sdiff_buffer[frame] = cap_buffer[frame * stride + left_idx ]
                          - cap_buffer[frame * stride + right_idx ];
    }
    didprint |= anachan(dsp, &sdiff_buffer[0], 1, 0,  "sdif", 0);
  }
  return didprint;
}

////////////////////////////////////////
// Read some data and maybe write checkfile
int tprops::read_stuff(alsa_dsp* dsp, datum* cap_buffer){

  int frames_read;

  for (;;){             // keep trying to read raw audio data
// DUBIOUS:  Should we do some better error checking here?
    frames_read = dsp->readi(&cap_buffer[0], block__fr);
    if (frames_read == block__fr) break;   // good!

// non-zero, but not as expected:
    if (frames_read == -EPIPE) {        // recover from overruns
      if (show["overrun"]) syslog(LOG_WARNING,
          "Read_stuff overrun: requested %i got %i: %s",
                block__fr, frames_read,
                snd_strerror(frames_read));
      snd_pcm_prepare(dsp->phndl);      // recover from overrun
      // fall through, to try reading again
    }
    else if (frames_read == -EAGAIN || frames_read == -EWOULDBLOCK) {
      return frames_read;
    }
    else if (frames_read > 0) {
      cerr << "ALSA capture: read_stuff: short read "
         << frames_read
         << " of " << block__fr
         << endl;
      // (presumably will be followed by EPIPE on next read)
      // fall through, to try reading again
    }
    else {    // some error we don't understand
      syslog(LOG_WARNING,
        "Peculiar error: ALSA capture: read_stuff requested %i got %i: %s",
              block__fr, frames_read, snd_strerror(frames_read));
      exeunt(1);
    }
  }
  // here with a buffer with the correct number of frames
  dsp->buf_serno++;

  return frames_read;
}

// write a buffer to the wfile
// header has already been written
void tprops::wfile_out(alsa_dsp* cap_dsp, datum* cap_buffer){
  if (wfile_count <= 0) {
    return;
  }
  int finbuf = block__fr;
  int do_frames = min(finbuf, wfile_count);
  int do_bytes = snd_pcm_frames_to_bytes(cap_dsp->phndl, do_frames);
  if (1) {
    int sts = write(wfile_fd, &cap_buffer[0], do_bytes);
    if (sts < 0) {
      syslog(LOG_ERR, "write error on raw output file (%i): %m",
                wfile_fd);
      exeunt(1);
    }
    if (sts != do_bytes) {
      syslog(LOG_ERR, "short write [%i] on raw output file (%i): %m",
                sts, wfile_fd);
      exeunt(1);
    }
    wfile_count -= do_frames;

    if (wfile_count <= 0) {            // check file is all written.
      close(wfile_fd);
      wfile_fd = -1;

// Tell the user the wfile is done:
      cerr << "<*> done with wfile: " << wfile_fname << endl;
    }
  }
}

int tprops::find_stuck(alsa_dsp* cap_dsp, datum* cap_buf){
  int didprint(0);
  if (!show["stuck"]) return didprint;
  if (getTime() - last_printout < snooze) return didprint;
  int finbuf = block__fr;
  int activeNsig = cap_dsp->channels.size();     // number of active signals
  typedef map<datum,int> counter;
  vector<counter> cnt(activeNsig);
  datum* cap_ptr(&cap_buf[0]);
  for (int ff = 0; ff < finbuf; ff++){
    for (int asig = 0; asig < activeNsig; asig++){
      datum value = cap_ptr[cap_dsp->channels[asig]];
      cnt[asig][value]++;
    }
    cap_ptr += cap_dsp->hwNchan;          // use a new frame next time
  }

// we now have a sorted list of values
// don't care how many times the value was seen.

  map<datum,int> histo;
  for (int asig = 0; asig < activeNsig; asig++) {
    counter::const_iterator getval = cnt[asig].begin();
    datum last = getval->first;
    getval++;

    for (; getval != cnt[asig].end(); getval++){
      datum value = getval->first;
      datum diff = value - last;
      histo[diff]++;
      last = value;
    }
  }

  map<datum,int>::const_iterator out = histo.begin();
  didprint = 1;
  cout << "........" << endl;
  for (int ii = 0; ii < 10; ii++) {
    if (out == histo.end()) break;
    cout << "rank: " << ii
        << "  delta: " << setw(6) << out->first
        << "  occurred: " << out-> second
        << endl;
    out++;
  }
  return didprint;
}

void tprops::krunch_entropy(alsa_dsp* cap_dsp, datum* cap_buf){
    unsigned char digest[20];
    int finbuf = block__fr;

    int hwNchan    = cap_dsp->hwNchan;          // channels the HW gives us,
                                                // whether active or not
    int do_out(sps && hwNchan && (rando_fd >= 0));
    int activeNsig = cap_dsp->channels.size();     // number of active signals
    datum* cap_ptr(&cap_buf[0]);
    vector<datum> chosen_buf(activeNsig);    // active channels only
    if (do_out) while (finbuf) {
      for (int asig = 0; asig < activeNsig; asig++){
        chosen_buf[asig] = cap_ptr[cap_dsp->channels[asig]];
      }

      SHA1_Update(&shacon, (unsigned char*) (&chosen_buf[0]),
                        activeNsig * sizeof(datum));
      cap_ptr += cap_dsp->hwNchan;          // use a new frame next time
      finbuf--;                         // fewer frames remaining
      sincon += activeNsig * sps;       // more entropy in shacon
      if (sincon > 170){
        SHA1_Final(digest, &shacon);
        chunky.Write(digest, sizeof(digest));
#if 0
        int sts = write(rando_fd, digest, sizeof(digest));
        if (sts != sizeof(digest)){
          syslog(LOG_ERR, "Write error on fifo (principal output): %m");
          sleep(10);            // or exit ???
        }
#endif
        SHA1_Init(&shacon);     // get ready for next time
        sincon = 0;
      }
    }

}

void captor::setup(){
// A small block size here seems to have little or no
// effect on efficiency.
// (In contrast, the output block size matters a lot.)
  tpx.block__fr = cap_dsp->intperiod__fr * 1;
  hwNchan    = cap_dsp->hwNchan;   // number of hw-supported channels

  ringlen = (cap_dsp->rate + tpx.block__fr - 1) / tpx.block__fr;
  ringptr = 0;
  cout << "ringlen: " << ringlen
        << "  total samples: " << ringlen * tpx.block__fr
        << "  versus: " << cap_dsp->rate
        << endl;
  cap_buf.resize(ringlen * tpx.block__fr * hwNchan);
  tpx.chunky.setup(tpx.bufpipe[1], tpx.obs);

  SHA1_Init(&tpx.shacon);
  tpx.sincon = 0;             // entropy in shacon
}

void captor::io_step(){
  datum* cap_base = &cap_buf[0];
  if (1) {
    if (ringptr > ringlen || ringptr < 0){
      cerr << "gack: broken ringptr" << endl;
      exit(1);
    }
    if (ringptr == ringlen) ringptr = 0;
    if (show["readcycle"]) cerr << ".";
    datum* cap_this = cap_base + ringptr*hwNchan;
    int finbuf = tpx.read_stuff(cap_dsp, cap_this);
    if (finbuf == -EAGAIN || finbuf == -EWOULDBLOCK) return;
    if (finbuf < 0) {
      cerr << "Read error should have been handled at lower level" << endl;
      exeunt(2);
    }
    if (finbuf != tpx.block__fr) {
      cerr << "wrong size read; should never happen" << endl;
      cerr << "finbuf: " << finbuf
                << "  block__fr: " << tpx.block__fr
                << endl;
      exeunt(2);
    }

// FIXME:  This seems to assume that the audio format e.g. S32_LE
// is the same as the local machine representation for numbers.
// Is there something along the lines of ntohl(3) that we
// should be using here?

    int didprint(0);
    if (cap_dsp->buf_serno >= 0) {
      didprint |= tpx.lockin(cap_dsp, cap_this);
      didprint |= tpx.analyze(cap_dsp, cap_this);
      tpx.krunch_entropy(cap_dsp, cap_this);
      didprint |= tpx.find_stuck(cap_dsp, cap_this);
      tpx.wfile_out(cap_dsp, cap_this);
    }
    if (didprint) tpx.last_printout = getTime();
    if (tpx.exit_flag && tpx.wfile_count <= 0) exeunt(0);
  }
  return;
}

////////////////////
// The whole idea of a pidfile is a crock.
// It is inconsistent with the idea that you might
// reasonably run multiple copies of turbid.
void write_pid_file(const string fname) {
  pidfile = "";                 // assume the worst, for now

  if (!fname.length()) return;
  FILE* junk = fopen(fname.c_str(), "r");
  if (junk) {
// ???!!! todo:  be smart if pidfile exists but points
// to extinct process ......
    syslog(LOG_WARNING, "PID file \"%s\" already exists, not overwriting",
        fname.c_str());
    fclose(junk);
    return;
  }

  FILE* p = fopen(fname.c_str(), "w");
  if (!p) {
    syslog(LOG_WARNING, "Couldn't open PID file \"%s\" for writing: %m.",
         fname.c_str());
  } else {
    int kidpid = getpid();
    fprintf(p, "Kidpid: %i\n", kidpid);
    fclose(p);
    pidfile = fname;    // remember fname in global (pidfile)
                        // for later use during cleanup
  }
}

void cleanup(int sts){
          for (string::size_type ii = 0;  ii < clean_me.size(); ii++) {
    int pid = clean_me[ii];
    if (pid > 0) kill(pid, SIGQUIT);
  }
  if (pidfile.length()) unlink(pidfile.c_str());
  decolorize(cout, colorized);
  exit(sts);
}

void trap_exit(int signum) {
  syslog(LOG_NOTICE, "%s stopping on signal %i", progname.c_str(), signum);
  cleanup(0);
}

void trap_detach(int signum) {
  syslog(LOG_NOTICE, "%s detaching on signal %i", progname.c_str(), signum);
  exit(0);
}

void dummy(int signum) {
  syslog(LOG_NOTICE, "Dummy caught signal %i", signum);
}

void exeunt(int const sts){
  if (daemonflag)
    syslog(LOG_NOTICE, "%s exiting, status %i", progname.c_str(), sts);
  cleanup(sts);
}
